from dataclasses import dataclass
from typing import Mapping


type operator = Add | Sub | Mult  # …


@dataclass
class Add:
    ...


@dataclass
class Sub:
    ...


@dataclass
class Mult:
    ...


type expr = Name | Constant | BinOp  # …


@dataclass
class Name:
    id: str


@dataclass
class Constant:
    value: int  # …


@dataclass
class BinOp:
    left: expr
    op: operator
    right: expr


def display_binop(op: operator) -> str:
    match op:
        case Add():
            return "Add()"
        case Sub():
            return "Sub()"
        case Mult():
            return "Mult()"


def display_inline(e: expr) -> str:
    match e:
        case Name(id):
            return "Name(id='" + id + "')"
        case Constant(value) if type(value) is int:
            return "Constant(value=" + str(value) + ")"
        case BinOp(left, op, right):
            return ("BinOp(left=" + display_inline(left) + ", "
                    + "op=" + display_binop(op) + ", "
                    + "right=" + display_inline(right) + ")")
        case _:
            raise ValueError("unsupported expression")


def display(e: expr, depth: int = 0, indent: int = 0) -> str:
    match e:
        case Name(id):
            return "\n" + depth * indent * " " + "Name(id='" + id + "')"
        case Constant(value) if type(value) is int:
            return "\n" + depth * indent * " " + "Constant(value=" + str(value) + ")"
        case BinOp(left, op, right):
            return ("\n" + depth * indent * " " + "BinOp(" +
                    "\n" + depth * indent * " " + "  left=" +
                    display(left, depth + 1, indent) + ", " +
                    "\n" + depth * indent * " " + "  op=" +
                    display_binop(op) + ", " +
                    "\n" + depth * indent * " " + "  right=" +
                    display(right, depth + 1, indent) + ")")
        case _:
            raise ValueError("unsupported expression")


def display_without_fields(e: expr, depth: int = 0) -> str:
    match e:
        case Name(id):
            return "\n" + depth * "    " + "Name('" + id + "')"
        case Constant(value) if type(value) is int:
            return "\n" + depth * "    " + "Constant(" + str(value) + ")"
        case BinOp(left, op, right):
            return ("\n" + depth * "    " + "BinOp(" +
                    display_without_fields(left, depth + 1) + ", " +
                    "\n" + (depth + 1) * "    " + display_binop(op) + ", " +
                    display_without_fields(right, depth + 1) + ")")
        case _:
            raise ValueError("unsupported expression")


def dump(e: expr, indent: int | None = None) -> str:
    if indent is None:
        return display_inline(e)
    else:
        return display(e, 0, indent)


def unparse_binop(op: operator) -> str:
    match op:
        case Add():
            return "+"
        case Sub():
            return "-"
        case Mult():
            return "*"


def unparse(e: expr) -> str:
    match e:
        case Name(id):
            return id
        case Constant(value) if type(value) is int:
            return str(value)
        case BinOp(left, op, right):
            return "(" + unparse(left) + " " + unparse_binop(op) + " " + unparse(right) + ")"
        case _:
            raise ValueError("unsupported expression")


def eval_binop(left_value: int, op: operator, right_value: int) -> int:
    match op:
        case Add():
            return left_value + right_value
        case Sub():
            return left_value - right_value
        case Mult():
            return left_value * right_value


def eval_expr(e: expr, m: Mapping[str, int]) -> int:
    match e:
        case Name(id):
            return m[id]
        case Constant(value) if type(value) is int:
            return value
        case BinOp(left, op, right):
            left_value: int = eval_expr(left, m)
            right_value: int = eval_expr(right, m)
            return eval_binop(left_value, op, right_value)
        case _:
            raise ValueError("unsupported expression")


e: expr = (
    BinOp(
        BinOp(
            Constant(3),
            Add(),
            BinOp(
                Name("x"),
                Mult(),
                Constant(5))),
        Mult(),
        BinOp(
            Name("y"),
            Add(),
            Constant(7)))
)


def main() -> None:
    print(e)
    print(display_inline(e))
    print(display(e, 0, 6))
    print(display_without_fields(e))
    print(dump(e))
    print(dump(e, indent=4))
    print(unparse(e))
    print(eval_expr(e, {"x": 5, "y": 9}))
    print(eval_expr(e, {"x": 6, "y": 2}))
    print(eval_expr(e, {"x": 0, "y": 0}))


if __name__ == '__main__':
    main()
